import base64
import binascii
import operator
import re
import traceback


from app.regex_base import REGEX_INSENSETIVE, REGEX_SENSETIVE
from app.utils import *


def score_with_regex(string):

    # Length Score
    # length = len(string)
    # if length > int(state.args.y) and length < int(state.args.s):
    #    localStringScores[string] += round(len(string) / 8, 2)
    # if length >= int(state.args.s):
    #    localStringScores[string] += 1

    # Reduction

    score = 0
    if ".." in string:
        score -= 5
    if "   " in string:
        score -= 5
    # Packer Strings
    if re.search(r"(WinRAR\\SFX)", string):
        score -= 4
    # US ASCII char
    if "\x1f" in string:
        score -= 4
    # Chains of 00s
    if string.count("0000000000") > 2:
        score -= 5
    # Repeated characters
    if re.search(r"(?!.* ([A-Fa-f0-9])\1{8,})", string):
        score -= 5

    # print("processing string: ", string)
    def filter_rg(string, regex_base, ignorecase):
        score_local = 0
        cats = ""
        flags = 0 if not ignorecase else re.IGNORECASE
        for cat, regexes in regex_base.items():
            found = False
            for regex in regexes:
                if m := re.search(regex[0], string, flags):
                    score_local += regex[1]
                    # print(cat, m)
                    found = True
            if found:
                cats += cat + ", "

        return score_local, cats

    cats = ""
    new_score, new_cats = filter_rg(string, REGEX_INSENSETIVE, True)
    score += new_score
    cats += new_cats
    new_score, new_cats = filter_rg(string, REGEX_SENSETIVE, False)
    score += new_score
    cats += new_cats

    return score, cats


def filter_string_set(string_set, state):
    # This is the only set we have - even if it's a weak one
    useful_set = []

    # Local string scores
    localStringScores = {}

    # Local UTF strings
    if getattr(state, "utf16strings", None) is None:
        state.utf16strings = []

    for string in string_set:

        # Goodware string marker
        goodstring = False
        goodcount = 0

        # Goodware Strings
        if string in state.good_strings_db:
            goodstring = True
            goodcount = state.good_strings_db[string]
            # print "%s - %s" % ( goodstring, good_strings[string] )
            if state.args.excludegood:
                continue

        # UTF
        original_string = string
        if string[:8] == "UTF16LE:":
            # print "removed UTF16LE from %s" % string
            string = string[8:]
            state.utf16strings.append(string)

        # Good string evaluation (after the UTF modification)
        if goodstring:
            # Reduce the score by the number of occurence in goodware files
            localStringScores[string] = (goodcount * -1) + 5
        else:
            localStringScores[string] = 0

        # PEStudio String Blacklist Evaluation
        if state.pestudio_available:
            (pescore, type) = get_pestudio_score(string, state.pestudio_strings)
            # print("PE Match: %s" % string)
            # Reset score of goodware files to 5 if blacklisted in PEStudio
            if type != "":
                state.pestudioMarker[string] = type
                # Modify the PEStudio blacklisted strings with their goodware stats count
                if goodstring:
                    pescore = pescore - (goodcount / 1000.0)
                    # print "%s - %s - %s" % (string, pescore, goodcount)
                localStringScores[string] = pescore

        if not goodstring:
            score, cats = score_with_regex(string)
            if state.args.trace:
                print(f"{string} - {score} - {cats}")
            localStringScores[string] += score
            state.string_to_comms[string] = cats
            # ENCODING DETECTIONS --------------------------------------------------
            try:
                if len(string) > 8:
                    # Try different ways - fuzz string
                    # Base64
                    if state.args.trace:
                        print("Starting Base64 string analysis ...")
                    for m_string in (
                        string,
                        string[1:],
                        string[:-1],
                        string[1:] + "=",
                        string + "=",
                        string + "==",
                    ):
                        if is_base_64(m_string):
                            try:
                                decoded_string = base64.b64decode(
                                    m_string, validate=False
                                )
                            except binascii.Error as e:
                                continue
                            if is_ascii_string(decoded_string, padding_allowed=True):
                                # print "match"
                                localStringScores[string] += 10
                                state.base64strings[string] = decoded_string
                    # Hex Encoded string
                    if state.args.trace:
                        print("Starting Hex encoded string analysis ...")
                    for m_string in [string, re.sub("[^a-zA-Z0-9]", "", string)]:
                        # print m_string
                        if is_hex_encoded(m_string):
                            # print("^ is HEX")
                            decoded_string = bytes.fromhex(m_string)
                            # print removeNonAsciiDrop(decoded_string)
                            if is_ascii_string(decoded_string, padding_allowed=True):
                                # not too many 00s
                                if "00" in m_string:
                                    if (
                                        len(m_string) / float(m_string.count("0"))
                                        <= 1.2
                                    ):
                                        continue
                                # print("^ is ASCII / WIDE")
                                localStringScores[string] += 8
                                state.hexEncStrings[string] = decoded_string
            except Exception as e:
                if state.args.debug:
                    traceback.print_exc()
                pass

            # Reversed String -----------------------------------------------------
            if string[::-1] in state.good_strings_db:
                localStringScores[string] += 10
                state.reversedStrings[string] = string[::-1]

            # Certain string reduce	-----------------------------------------------
            if re.search(r"(rundll32\.exe$|kernel\.dll$)", string, re.IGNORECASE):
                localStringScores[string] -= 4

        # Set the global string score
        state.stringScores[original_string] = localStringScores[string]

        if state.args.debug:
            if string in state.utf16strings:
                is_utf = True
            else:
                is_utf = False
                # print "SCORE: %s\tUTF: %s\tSTRING: %s" % ( localStringScores[string], is_utf, string )

    sorted_set = sorted(
        localStringScores.items(), key=operator.itemgetter(1), reverse=True
    )

    # Only the top X strings
    c = 0
    result_set = []
    for string in sorted_set:

        # Skip the one with a score lower than -z X
        if not state.args.noscorefilter and not state.args.inverse:
            if string[1] < int(state.args.z):
                continue

        if string[0] in state.utf16strings:
            result_set.append("UTF16LE:%s" % string[0])
        else:
            result_set.append(string[0])

        # c += 1
        # if c > int(state.args.rc):
        #    break

    if state.args.trace:
        print("RESULT SET:")
        print(result_set)

    # return the filtered set
    return result_set


def filter_opcode_set(state, opcode_set: list[str], good_opcodes_db) -> list[str]:
    # Preferred Opcodes
    pref_opcodes = [" 34 ", "ff ff ff "]

    # Useful set
    useful_set = []
    pref_set = []

    for opcode in opcode_set:
        # Exclude all opcodes found in goodware
        if opcode in good_opcodes_db:
            if state.args.trace:
                print("skipping %s" % opcode)
            continue

        # Format the opcode
        formatted_opcode = get_opcode_string(opcode)

        # Preferred opcodes
        set_in_pref = False
        for pref in pref_opcodes:
            if pref in formatted_opcode:
                pref_set.append(formatted_opcode)
                set_in_pref = True
        if set_in_pref:
            continue

        # Else add to useful set
        useful_set.append(get_opcode_string(opcode))

    # Preferred opcodes first
    useful_set = pref_set + useful_set

    # Only return the number of opcodes defined with the "-n" parameter
    return useful_set[: int(state.args.n)]


def sample_string_evaluation(
    string_stats, opcode_stats, file_info, state, utf16string_stats
):
    # Generate Stats -----------------------------------------------------------
    print("[+] Generating statistical data ...")
    file_strings = {}
    file_utf16strings = {}
    file_opcodes = {}
    combinations = {}
    inverse_stats = {}
    max_combi_count = 0
    super_rules = []

    def extract_stats_by_file(stats, outer_dict, flt=lambda x: x):
        for token, value in stats.items():
            # print(value)
            count = 0
            files = []
            if type(value) == dict:
                count = value["count"]
                files = value["files"]
            else:
                count = value.count
                files = value.files
            if flt(count):
                for filePath in files:
                    if filePath in outer_dict:
                        outer_dict[filePath].append(token)
                    else:
                        outer_dict[filePath] = [token]

    # OPCODE EVALUATION -----------------------------------------------
    extract_stats_by_file(opcode_stats, file_opcodes, lambda x: x < 10)

    # STRING EVALUATION -------------------------------------------------------

    # Iterate through strings found in malware files

    extract_stats_by_file(string_stats, file_strings)
    extract_stats_by_file(utf16string_stats, file_utf16strings)
    if not state.args.nosuper:
        for string in string_stats:
            # SUPER RULE GENERATION -----------------------------------------------
            # SUPER RULES GENERATOR	- preliminary work
            # If a string occurs more than once in different files
            # print sample_string_stats[string].count
            if len(string_stats[string].files) > 1:
                if state.args.debug:
                    print(
                        'OVERLAP Count: %s\nString: "%s"%s'
                        % (
                            string_stats[string].count,
                            string,
                            "\nFILE: ".join(string_stats[string].files),
                        )
                    )
                # Create a combination string from the file set that matches to that string
                combi = ":".join(sorted(string_stats[string].files))
                # print "STRING: " + string
                if state.args.debug:
                    print("COMBI: " + combi)
                # If combination not yet known
                if combi not in combinations:
                    combinations[combi] = {}
                    combinations[combi]["count"] = 1
                    combinations[combi]["strings"] = []
                    combinations[combi]["strings"].append(string)
                    combinations[combi]["files"] = string_stats[string].files
                else:
                    combinations[combi]["count"] += 1
                    combinations[combi]["strings"].append(string)
                # Set the maximum combination count
                max_combi_count = (
                    combinations[combi]["count"]
                    if combinations[combi]["count"] > max_combi_count
                    else max_combi_count
                )
                # print "Max Combi Count set to: %s" % max_combi_count

    print("[+] Generating Super Rules ... (a lot of magic)")
    for combi_count in range(max_combi_count, 1, -1):
        for combi in combinations:
            if combi_count == combinations[combi]["count"]:
                # print "Count %s - Combi %s" % ( str(combinations[combi]["count"]), combi )
                # Filter the string set
                # print "BEFORE"
                # print len(combinations[combi]["strings"])
                # print combinations[combi]["strings"]
                string_set = combinations[combi]["strings"]
                combinations[combi]["strings"] = []
                combinations[combi]["strings"] = filter_string_set(string_set, state)
                # print combinations[combi]["strings"]
                # print "AFTER"
                # print len(combinations[combi]["strings"])
                # Combi String count after filtering
                # print "String count after filtering: %s" % str(len(combinations[combi]["strings"]))

                # If the string set of the combination has a required size
                if len(combinations[combi]["strings"]) >= int(state.args.w):
                    # Remove the files in the combi rule from the simple set
                    if state.args.nosimple:
                        for file in combinations[combi]["files"]:
                            if file in file_strings:
                                del file_strings[file]
                    # Add it as a super rule
                    print(
                        "[-] Adding Super Rule with %s strings."
                        % str(len(combinations[combi]["strings"]))
                    )
                    # if state.args.debug:
                    # print "Rule Combi: %s" % combi
                    super_rules.append(combinations[combi])

    # Return all data
    return (file_strings, file_opcodes, combinations, super_rules, inverse_stats)



#!/usr/bin/env python

__version__ = "0.24.0"


import os
import sys

import traceback
import time
from collections import Counter
import signal as signal_module

from app.args import get_args
from app.utils import load
from app.parse_files import parse_good_dir, processSampleDir
from app.utils import (
    emptyFolder,
    get_abs_path,
    getIdentifier,
    getPrefix,
    getReference,
    initialize_pestudio_strings,
    save,
)
from app.config import DB_PATH


def load_db(file, local_counter, prefix):
    if file.startswith(prefix):
        try:
            filePath = os.path.join(DB_PATH, file)
            print("[+] Loading %s ..." % filePath)
            js = load(get_abs_path(filePath))
            local_counter.update(js)
            print(
                "[+] Total: %s / Added %d entries"
                % (len(local_counter), len(local_counter) - strings_num)
            )
        except Exception as e:
            traceback.print_exc()
    return len(local_counter)


class State:
    def __init__(
        self,
        args,
        good_strings_db,
        good_opcodes_db,
        good_imphashes_db,
        good_exports_db,
        pestudio_available,
        pestudio_strings,
    ):
        self.base64strings = {}
        self.reversedStrings = {}
        self.hexEncStrings = {}
        self.pestudioMarker = {}
        self.stringScores = {}
        self.good_strings_db = (good_strings_db,)
        self.good_opcodes_db = good_opcodes_db
        self.good_imphashes_db = good_imphashes_db
        self.good_exports_db = good_exports_db
        self.pestudio_available = pestudio_available
        self.pestudio_strings = pestudio_strings
        self.args = args
        self.string_to_comms = dict()


# CTRL+C Handler --------------------------------------------------------------
def signal_handler(signal_name, frame):
    print("> yarGen's work has been interrupted")
    sys.exit(0)


def print_welcome():
    print("------------------------------------------------------------------------")
    print("                   _____            ")
    print("    __ _____ _____/ ___/__ ___      ")
    print("   / // / _ `/ __/ (_ / -_) _ \\     ")
    print("   \\_, /\\_,_/_/  \\___/\\__/_//_/     ")
    print("  /___/  Yara Rule Generator        ")
    print("         Florian Roth, August 2023, Version %s" % __version__)
    print("   ")
    print("  Note: Rules have to be post-processed")
    print("  See this post for details: https://medium.com/@cyb3rops/121d29322282")
    print("------------------------------------------------------------------------")


# MAIN ################################################################
if __name__ == "__main__":
    print_welcome()

    # Signal handler for CTRL+C
    signal_module.signal(signal_module.SIGINT, signal_handler)
    args = get_args()
    # Read PEStudio string list
    pestudio_strings = {}
    pestudio_available = False

    # Identifier
    sourcepath = args.m
    if args.g:
        sourcepath = args.g
    args.identifier = getIdentifier(args.b, sourcepath)
    print("[+] Using identifier '%s'" % args.identifier)

    # Reference
    args.reference = getReference(args.r)
    print("[+] Using reference '%s'" % args.reference)

    # Prefix
    args.prefix = getPrefix(args.p, args.identifier)
    print("[+] Using prefix '%s'" % args.prefix)

    if strs := initialize_pestudio_strings():
        pestudio_available = True
        pestudio_strings = strs

    # Highly specific string score
    args.score_highly_specific = int(args.x)

    # Scan goodware files
    if args.g:
        print("[+] Processing goodware files ...")
        good_strings_db, good_opcodes_db, good_imphashes_db, good_exports_db = (
            parse_good_dir(args, args.g, args.nr, args.oe)
        )

        # Update existing databases
        if args.u:
            try:
                print("[+] Updating databases ...")

                # Evaluate the database identifiers
                db_identifier = ""
                if args.i != "":
                    db_identifier = "-%s" % args.i
                strings_db = "./dbs/good-strings%s.db" % db_identifier
                opcodes_db = "./dbs/good-opcodes%s.db" % db_identifier
                imphashes_db = "./dbs/good-imphashes%s.db" % db_identifier
                exports_db = "./dbs/good-exports%s.db" % db_identifier

                # Strings -----------------------------------------------------
                print("[+] Updating %s ..." % strings_db)
                good_pickle = load(get_abs_path(strings_db))
                print("Old string database entries: %s" % len(good_pickle))
                good_pickle.update(good_strings_db)
                print("New string database entries: %s" % len(good_pickle))
                save(good_pickle, strings_db)

                # Opcodes -----------------------------------------------------
                print("[+] Updating %s ..." % opcodes_db)
                good_opcode_pickle = load(get_abs_path(opcodes_db))
                print("Old opcode database entries: %s" % len(good_opcode_pickle))
                good_opcode_pickle.update(good_opcodes_db)
                print("New opcode database entries: %s" % len(good_opcode_pickle))
                save(good_opcode_pickle, opcodes_db)

                # Imphashes ---------------------------------------------------
                print("[+] Updating %s ..." % imphashes_db)
                good_imphashes_pickle = load(get_abs_path(imphashes_db))
                print("Old opcode database entries: %s" % len(good_imphashes_pickle))
                good_imphashes_pickle.update(good_imphashes_db)
                print("New opcode database entries: %s" % len(good_imphashes_pickle))
                save(good_imphashes_pickle, imphashes_db)

                # Exports -----------------------------------------------------
                print("[+] Updating %s ..." % exports_db)
                good_exports_pickle = load(get_abs_path(exports_db))
                print("Old opcode database entries: %s" % len(good_exports_pickle))
                good_exports_pickle.update(good_exports_db)
                print("New opcode database entries: %s" % len(good_exports_pickle))
                save(good_exports_pickle, exports_db)

            except Exception as e:
                traceback.print_exc()

        # Create new databases
        if args.c:
            print("[+] Creating local database ...")
            # Evaluate the database identifiers
            db_identifier = ""
            if args.i != "":
                db_identifier = "-%s" % args.i
            strings_db = "./dbs/good-strings%s.db" % db_identifier
            opcodes_db = "./dbs/good-opcodes%s.db" % db_identifier
            imphashes_db = "./dbs/good-imphashes%s.db" % db_identifier
            exports_db = "./dbs/good-exports%s.db" % db_identifier

            # Creating the databases
            print(
                "[+] Using '%s' as filename for newly created strings database"
                % strings_db
            )
            print(
                "[+] Using '%s' as filename for newly created opcodes database"
                % opcodes_db
            )
            print(
                "[+] Using '%s' as filename for newly created opcodes database"
                % imphashes_db
            )
            print(
                "[+] Using '%s' as filename for newly created opcodes database"
                % exports_db
            )

            try:
                for file in [strings_db, opcodes_db, imphashes_db, exports_db]:

                    if os.path.isfile(file):
                        input(
                            "File %s alread exists. Press enter to proceed or CTRL+C to exit."
                            % file
                        )
                        os.remove(file)

                # Strings
                good_json = Counter()
                good_json = good_strings_db
                # Opcodes
                good_op_json = Counter()
                good_op_json = good_opcodes_db
                # Imphashes
                good_imphashes_json = Counter()
                good_imphashes_json = good_imphashes_db
                # Exports
                good_exports_json = Counter()
                good_exports_json = good_exports_db

                # Save
                save(good_json, strings_db)
                save(good_op_json, opcodes_db)
                save(good_imphashes_json, imphashes_db)
                save(good_exports_json, exports_db)

                print(
                    "New database with %d string, %d opcode, %d imphash, %d export entries created. "
                    "(remember to use --opcodes to extract opcodes from the samples and create the opcode databases)"
                    % (
                        len(good_strings_db),
                        len(good_opcodes_db),
                        len(good_imphashes_db),
                        len(good_exports_db),
                    )
                )
            except Exception as e:
                traceback.print_exc()

    # Analyse malware samples and create rules
    else:
        print("[+] Reading goodware strings from database 'good-strings.db' ...")
        print(
            "    (This could take some time and uses several Gigabytes of RAM depending on your db size)"
        )

        good_strings_db = Counter()
        good_opcodes_db = Counter()
        good_imphashes_db = Counter()
        good_exports_db = Counter()

        opcodes_num = 0
        strings_num = 0
        imphash_num = 0
        exports_num = 0

        # Initialize all databases
        for file in os.listdir(get_abs_path(DB_PATH)):
            if not file.endswith(".db"):
                continue  # String databases
            strings_num = load_db(file, good_strings_db, "good-strings")
            opcodes_num = load_db(file, good_opcodes_db, "good-opcodes")
            imphash_num = load_db(file, good_imphashes_db, "good-imphash")
            exports_num = load_db(file, good_exports_db, "good-exports")

        if args.opcodes and len(good_opcodes_db) < 1:
            print(
                "[E] Missing goodware opcode databases."
                "    Please run 'yarGen.py --update' to retrieve the newest database set."
            )
            args.opcodes = False

        if len(good_exports_db) < 1 and len(good_imphashes_db) < 1:
            print(
                "[E] Missing goodware imphash/export databases. "
                "    Please run 'yarGen.py --update' to retrieve the newest database set."
            )

        if len(good_strings_db) < 1 and not args.c:
            print(
                "[E] Error - no goodware databases found. "
                "    Please run 'yarGen.py --update' to retrieve the newest database set."
            )
            sys.exit(1)

    # If malware directory given
    if args.m:
        # Deactivate super rule generation if there's only a single file in the folder
        if len(os.listdir(args.m)) < 2:
            args.nosuper = True

        # Special strings
        state = State(
            args,
            good_strings_db,
            good_opcodes_db,
            good_imphashes_db,
            good_exports_db,
            pestudio_available,
            pestudio_strings,
        )
        # Dropzone mode
        if args.dropzone:
            # Monitoring folder for changes
            print(
                "Monitoring %s for new sample files (processed samples will be removed)"
                % args.m
            )
            while True:
                if len(os.listdir(args.m)) > 0:
                    # Deactivate super rule generation if there's only a single file in the folder
                    if len(os.listdir(args.m)) < 2:
                        args.nosuper = True
                    else:
                        args.nosuper = False
                    # Read a new identifier
                    identifier = getIdentifier(args.b, args.m)
                    # Read a new reference
                    reference = getReference(args.r)
                    # Generate a new description prefix
                    prefix = getPrefix(args.p, identifier)
                    # Process the samples
                    processSampleDir(args.m, state)
                    # Delete all samples from the dropzone folder
                    emptyFolder(args.m)
                time.sleep(1)
        else:
            # Scan malware files
            print("[+] Processing malware files ...")
            processSampleDir(args.m, state)

        print("[+] yarGen run finished")




REGEX_INSENSETIVE = {
    # "packers": [(r'(WinRAR\\SFX)', -4)],
    # "repeated_chars": [(r'(?!.* ([A-Fa-f0-9])\1{8,})', -5)],
    "drives": [(r"[A-Za-z]:\\", -4)],
    "exe_extensions": [
        (
            r"(\.exe|\.pdb|\.scr|\.log|\.cfg|\.txt|\.dat|\.msi|\.com|\.bat|\.dll|\.pdb|\.vbs|\.tmp|\.sys|\.ps1|\.vbp|\.hta|\.lnk)",
            4,
        )
    ],
    "system_keywords": [
        (
            r"(cmd.exe|system32|users|Documents and|SystemRoot|Grant|hello|password|process|log)",
            5,
        )
    ],
    "protocol_keywords": [(r"(ftp|irc|smtp|command|GET|POST|Agent|tor2web|HEAD)", 5)],
    "connection_keywords": [(r"(error|http|closed|fail|version|proxy)", 3)],
    "temp_and_recycler": [(r"(TEMP|Temporary|Appdata|Recycler)", 4)],
    "malicious_keywords": [
        (
            r"(scan|sniff|poison|intercept|fake|spoof|sweep|dump|flood|inject|forward|scan|vulnerable|credentials|creds|coded|p0c|Content|host)",
            5,
        )
    ],
    "network_keywords": [
        (
            r"(address|port|listen|remote|local|process|service|mutex|pipe|frame|key|lookup|connection)",
            3,
        )
    ],
    "drive": [(r"([C-Zc-z]:\\)", 4)],
    "IP": [
        (
            r"\b(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\b",
            5,
        )
    ],
    "copyright": [(r"(coded | c0d3d |cr3w\b|Coded by |codedby)", 7)],
    "extensions_generic": [(r"\.[a-zA-Z]{3}\b", 3)],
    "alll_characters": [(r"^[A-Z][a-z]{5,}$", 2)],
    "URL": [(r"(%[a-z][:\-,;]|\\\\%s|\\\\[A-Z0-9a-z%]+\\[A-Z0-9a-z%]+)", 2.5)],
    "certificates": [
        (r"(thawte|trustcenter|signing|class|crl|certificate|assembly)", -4)
    ],
    "parameters": [(r"( \-[a-z]{,2}[\s]?[0-9]?| /[a-z]+[\s]?[\w]*)", 4)],
    "directory": [(r"([a-zA-Z]:|^|%)\\[A-Za-z]{4,30}\\", 4)],
    "executable_no_dir": [(r"^[^\\]+\.(exe|com|scr|bat|sys)$", 4)],
    "date_placeholders": [(r"(yyyy|hh:mm|dd/mm|mm/dd|%s:%s:)", 3)],
    "placeholders": [(r"[^A-Za-z](%s|%d|%i|%02d|%04d|%2d|%3s)[^A-Za-z]", 3)],
    "string_parts": [
        (r"(cmd|com|pipe|tmp|temp|recycle|bin|secret|private|AppData|driver|config)", 3)
    ],
    "programming": [
        (
            r"(execute|run|system|shell|root|cimv2|login|exec|stdin|read|process|netuse|script|share)",
            3,
        )
    ],
    "credentials": [
        (
            r"(user|pass|login|logon|token|cookie|creds|hash|ticket|NTLM|LMHASH|kerberos|spnego|session|identif|account|login|auth|privilege)",
            3,
        )
    ],
    "malware": [(r"(\.[a-z]/[^/]+\.txt)", 3)],
    "variables": [(r"%[A-Z_]+%", 4)],
    "RATs": [
        (
            r"(spy|logger|dark|cryptor|RAT\b|eye|comet|evil|xtreme|poison|meterpreter|metasploit|/veil|Blood)",
            5,
        )
    ],
    "paths": [(r"^[Cc]:\\\\[^PW]", 3)],
    "missed_user_profiles": [
        (
            r"[\\](users|profiles|username|benutzer|Documents and Settings|Utilisateurs|Utenti|Usuários)[\\]",
            3,
        )
    ],
    "strings_with_numbers": [(r"^[A-Z][a-z]+[0-9]+$", 1)],
    "spying": [(r"(implant)", 1)],
    "special_strings": [(r"(\\\\\.\\|kernel|.dll|usage|\\DosDevices\\)", 5)],
    "parameters": [(r"( \-[a-z] | /[a-z] | \-[a-z]:[a-zA-Z]| \/[a-z]:[a-zA-Z])", 4)],
    "file": [(r"^[a-zA-Z0-9]{3,40}\.[a-zA-Z]{3}", 3)],
    "comment": [(r"^([\*\#]+ |\[[\*\-\+]\] |[\-=]> |\[[A-Za-z]\] )", 4)],
    "typo": [(r"(!\.$|!!!$| :\)$| ;\)$|fucked|[\w]\.\.\.\.$)", 4)],
    "base64": [
        (r"^(?:[A-Za-z0-9+/]{4}){30,}(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$", 7)
    ],
    "b64exec": [
        (
            r"(TVqQAAMAAAAEAAAA//8AALgAAAA|TVpQAAIAAAAEAA8A//8AALgAAAA|TVqAAAEAAAAEABAAAAAAAAAAAAA|TVoAAAAAAAAAAAAAAAAAAAAAAAA|TVpTAQEAAAAEAAAA//8AALgAAAA)",
            5,
        )
    ],
    "malicious_intent": [
        (
            r"(loader|cmdline|ntlmhash|lmhash|infect|encrypt|exec|elevat|dump|target|victim|override|traverse|mutex|pawnde|exploited|shellcode|injected|spoofed|dllinjec|exeinj|reflective|payload|inject|back conn)",
            5,
        )
    ],
    "privilege": [
        (r"(administrator|highest|system|debug|dbg|admin|adm|root) privelege", 4)
    ],
    "system_file": [(r"(LSASS|SAM|lsass.exe|cmd.exe|LSASRV.DLL)", 4)],
    "compiler": [(r"(Release|Debug|bin|sbin)", 2)],
    "pe_exe": [(r"(\.exe|\.dll|\.sys)$", 4)],
    "string_valid": [(r"(^\\\\)", 1)],
    "malware_related": [
        (
            r"(Management Support Team1|/c rundll32|DTOPTOOLZ Co.|net start|Exec|taskkill)",
            4,
        )
    ],
    "powershell": [
        (
            r"(bypass|windowstyle | hidden |-command|IEX |Invoke-Expression|Net.Webclient|Invoke[A-Z]|\
                    Net.WebClient|-w hidden |-encoded-encodedcommand| -nop |MemoryLoadLibrary|FromBase64String|Download|EncodedCommand)",
            4,
        )
    ],
    "wmi": [(r"( /c WMIC)", 3)],
    "windows_commands": [
        (
            r"( net user | net group |ping |whoami |bitsadmin |rundll32.exe javascript:|\
                            schtasks.exe /create|/c start )",
            3,
        )
    ],
    "signing_certificates": [(r"( Inc | Co.|  Ltd.,| LLC| Limited)", 2)],
    "privilege_escalation": [(r"(sysprep|cryptbase|secur32)", 2)],
    "webshells": [(r"(isset\($post\[|isset\($get\[|eval\(Request)", 2)],
    "suspicious_words": [
        (
            r"(impersonate|drop|upload|download|execute|shell|\bcmd\b|decode|rot13|decrypt)",
            2,
        )
    ],
    "suspicious_words2": [
        (
            r"([+] |[-] |[*] |injecting|exploit|dumped|dumping|scanning|scanned|elevation|\
                            elevated|payload|vulnerable|payload|reverse connect|bind shell|reverse shell| dump |\
                            back connect |privesc|privilege escalat|debug privilege| inject |interactive shell|\
                            shell commands| spawning |] target |] Transmi|] Connect|] connect|] Dump|] command |\
                            ] token|] Token |] Firing | hashes | etc/passwd| SAM | NTML|unsupported target|\
                            race condition|Token system |LoaderConfig| add user |ile upload |ile download |\
                            Attaching to |ser has been successfully added|target system |LSA Secrets|DefaultPassword|\
                            Password: |loading dll|.Execute\(|Shellcode|Loader|inject x86|inject x64|bypass|katz|\
                            spoit|ms[0-9][0-9][^0-9]|\bCVE[^a-zA-Z]|privilege::|lsadump|door)",
            4,
        )
    ],
    "mutex_pipes": [(r"(Mutex|NamedPipe|\\Global\\|\\pipe\\)", 3)],
    "usage": [(r"(isset\($post\[|isset\($get\[)", 2)],
    "hash": [(r"\b([a-f0-9]{32}|[a-f0-9]{40}|[a-f0-9]{64})\b", 2)],
    "persistence": [(r"(sc.exe |schtasks|at \\\\|at [0-9]{2}:[0-9]{2})", 3)],
    "unix": [
        (
            r"(;chmod |; chmod |sh -c|/dev/tcp/|/bin/telnet|selinux| shell| cp /bin/sh )",
            3,
        )
    ],
    "attack": [
        (
            r"(attacker|brute force|bruteforce|connecting back|EXHAUSTIVE|exhaustion| spawn| evil| elevated)",
            3,
        )
    ],
    "less_value": [
        (r"(abcdefghijklmnopqsst|ABCDEFGHIJKLMNOPQRSTUVWXYZ|0123456789:;)", -5)
    ],
    "vb_backdoors": [(r"(kill|wscript|plugins|svr32|Select )", 3)],
    "susp_strings_combo": [(r"([a-z]{4,}[!\?]|\[[!+\-]\] )", 3)],
    "special_chars": [(r"(-->|!!!| <<< | >>> )", 5)],
    "swear": [(r"\b(fuck|damn|shit|penis|nigger)\b", 5)],
    "scripts": [
        (r"(%APPDATA%|%USERPROFILE%|Public|Roaming|& del|& rm| && |script)", 3)
    ],
    "uacme": [(r"(Elevation|pwnd|pawn|elevate to)", 3)],
}

REGEX_SENSETIVE = {
    "all_caps": [(r"^[A-Z]{6,}$", 3)],
    "all_lower": [(r"^[a-z]{6,}$", 3)],
    "all_lower_with_space": [(r"^[a-z\s]{6,}$", 2)],
    "javascript": [
        (
            r'(new ActiveXObject\("WScript.Shell"\).Run|.Run\("cmd.exe|.Run\("%comspec%\)|.Run\("c:\\Windows|.RegisterXLL\()',
            3,
        )
    ],
    "ua_keywords": [
        (r"(Mozilla|MSIE|Windows NT|Macintosh|Gecko|Opera|User\-Agent)", 5)
    ],
}



import binascii
from collections import Counter
import os
import traceback
from dataclasses import dataclass

from hashlib import sha256
from typing import Self, Tuple
from app.rule_generator import generate_rules
from app.scoring import sample_string_evaluation

from app.utils import extract_opcodes, get_files, get_pe_info
import yargen_rs


@dataclass
class StringInfo:
    def __init__(
        self,
        count: int,
        is_utf16: bool = False,
        files: list[str] = [],
    ):
        self.count = count
        self.is_utf16 = is_utf16
        self.files = files

    def __str__(self):
        if self.is_utf16:
            return "UTF16: %s" % self.count
        else:
            return "%s" % self.count


def extract_strings(fileData):
    """Use Rust implementation for better performance"""
    strings = {
        s[0]: StringInfo(s[1])
        for s in yargen_rs.extract_strings(fileData, 5, 128, False)
    }
    utf16_strings = {
        s[0]: StringInfo(s[1], True)
        for s in yargen_rs.extract_strings(fileData, 5, 128, True)
    }
    return strings, utf16_strings


RELEVANT_EXTENSIONS = [
    ".asp",
    ".vbs",
    ".ps",
    ".ps1",
    ".tmp",
    ".bas",
    ".bat",
    ".cmd",
    ".com",
    ".cpl",
    ".crt",
    ".dll",
    ".exe",
    ".msc",
    ".scr",
    ".sys",
    ".vb",
    ".vbe",
    ".vbs",
    ".wsc",
    ".wsf",
    ".wsh",
    ".input",
    ".war",
    ".jsp",
    ".php",
    ".asp",
    ".aspx",
    ".psd1",
    ".psm1",
    ".py",
]


def parse_sample_dir(
    dir, state, notRecursive=False, generateInfo=False, onlyRelevantExtensions=False
):
    # Prepare dictionary
    string_stats = {}
    utf16string_stats = {}
    opcode_stats = {}
    file_info = {}
    known_sha1sums = []

    for filePath in get_files(dir, notRecursive):
        try:
            print("[+] Processing %s ..." % filePath)

            # Get Extension
            extension = os.path.splitext(filePath)[1].lower()
            if not extension in RELEVANT_EXTENSIONS and onlyRelevantExtensions:
                if state.args.debug:
                    print("[-] EXTENSION %s - Skipping file %s" % (extension, filePath))
                continue

            # Info file check
            if os.path.basename(filePath) == os.path.basename(
                state.args.b
            ) or os.path.basename(filePath) == os.path.basename(state.args.r):
                continue

            # Size Check
            size = 0
            try:
                size = os.stat(filePath).st_size
                if size > (state.args.fs * 1024 * 1024):
                    if state.args.debug:
                        print(
                            "[-] File is to big - Skipping file %s (use -fs to adjust this behaviour)"
                            % (filePath)
                        )
                    continue
            except Exception as e:
                pass

            # Check and read file
            try:
                with open(filePath, "rb") as f:
                    fileData = f.read()
            except Exception as e:
                print("[-] Cannot read file - skipping %s" % filePath)

            # Extract strings from file
            strings, utf16strings = extract_strings(fileData)
            for s in strings:
                strings[s].files = set(set(strings[s].files) | {filePath})
            for s in utf16strings:
                utf16strings[s].files = set(set(utf16strings[s].files) | {filePath})
            # print(strings, utf16strings)
            # Extract opcodes from file
            opcodes = []
            if state.args.opcodes:
                print("[-] Extracting OpCodes: %s" % filePath)
                opcodes = extract_opcodes(fileData)

            # Add sha256 value
            if generateInfo:
                sha256sum = sha256(fileData).hexdigest()
                file_info[filePath] = {}
                file_info[filePath]["hash"] = sha256sum
                file_info[filePath]["imphash"], file_info[filePath]["exports"] = (
                    get_pe_info(fileData)
                )

            # Skip if hash already known - avoid duplicate files
            if sha256sum in known_sha1sums:
                # if state.args.debug:
                print(
                    "[-] Skipping strings/opcodes from %s due to MD5 duplicate detection"
                    % filePath
                )
                continue
            else:
                known_sha1sums.append(sha256sum)

            # Magic evaluation
            if not state.args.nomagic:
                file_info[filePath]["magic"] = binascii.hexlify(fileData[:2]).decode(
                    "ascii"
                )
            else:
                file_info[filePath]["magic"] = ""

            # File Size
            file_info[filePath]["size"] = os.stat(filePath).st_size

            # Add stats for basename (needed for inverse rule generation)
            fileName = os.path.basename(filePath)
            folderName = os.path.basename(os.path.dirname(filePath))
            if fileName not in file_info:
                file_info[fileName] = {}
                file_info[fileName]["count"] = 0
                file_info[fileName]["hashes"] = []
                file_info[fileName]["folder_names"] = []
            file_info[fileName]["count"] += 1
            file_info[fileName]["hashes"].append(sha256sum)
            if folderName not in file_info[fileName]["folder_names"]:
                file_info[fileName]["folder_names"].append(folderName)

            def merge_stats(new_stats, old_stats):
                for string, info in new_stats.items():
                    if string not in old_stats:
                        old_stats[string] = info

                    elif info.is_utf16 == old_stats[string].is_utf16:
                        old_stats[string].count += new_stats[string].count
                        for f in new_stats[string].files:
                            if f not in old_stats[string].files:
                                old_stats[string].files.append(f)
                    else:
                        raise ValueError("String %s has different encoding" % string)

            merge_stats(strings, string_stats)
            merge_stats(utf16strings, utf16string_stats)

            # Add opcodes to statistics
            for opcode in opcodes:
                # Opcode is not already known
                if opcode not in opcode_stats:
                    opcode_stats[opcode] = {}
                    opcode_stats[opcode]["count"] = 0
                    opcode_stats[opcode]["files"] = []
                    opcode_stats[opcode]["files_basename"] = {}
                # Opcode count
                opcode_stats[opcode]["count"] += 1
                # Add file information
                if fileName not in opcode_stats[opcode]["files_basename"]:
                    opcode_stats[opcode]["files_basename"][fileName] = 0
                opcode_stats[opcode]["files_basename"][fileName] += 1
                if filePath not in opcode_stats[opcode]["files"]:
                    opcode_stats[opcode]["files"].append(filePath)

            if state.args.debug:
                print(
                    "[+] Processed "
                    + filePath
                    + " Size: "
                    + str(size)
                    + " Strings: "
                    + str(len(string_stats))
                    + " OpCodes: "
                    + str(len(opcode_stats))
                    + " ... "
                )

        except Exception as e:
            traceback.print_exc()
            print("[E] ERROR reading file: %s" % filePath)

    return string_stats, opcode_stats, file_info, utf16string_stats


def parse_good_dir(state, dir, notRecursive=False, onlyRelevantExtensions=True):
    # Prepare dictionary
    all_strings = Counter()
    all_opcodes = Counter()
    all_imphashes = Counter()
    all_exports = Counter()

    for filePath in get_files(dir, notRecursive):
        # Get Extension
        extension = os.path.splitext(filePath)[1].lower()
        if extension not in RELEVANT_EXTENSIONS and onlyRelevantExtensions:
            if state.args.debug:
                print("[-] EXTENSION %s - Skipping file %s" % (extension, filePath))
            continue

        # Size Check
        size = 0
        try:
            size = os.stat(filePath).st_size
            if size > (state.args.fs * 1024 * 1024):
                continue
        except Exception as e:
            pass

        # Check and read file
        try:
            with open(filePath, "rb") as f:
                fileData = f.read()
        except Exception as e:
            print("[-] Cannot read file - skipping %s" % filePath)

        # Extract strings from file
        strings = extract_strings(state.args.s, fileData)
        # Append to all strings
        all_strings.update(strings)

        # Extract Opcodes from file
        opcodes = []
        if state.args.opcodes:
            print("[-] Extracting OpCodes: %s" % filePath)
            opcodes = extract_opcodes(fileData)
            # Append to all opcodes
            all_opcodes.update(opcodes)

        # Imphash and Exports
        (imphash, exports) = get_pe_info(fileData)
        if imphash != "":
            all_imphashes.update([imphash])
        all_exports.update(exports)
        if state.args.debug:
            print(
                "[+] Processed %s - %d strings %d opcodes %d exports and imphash %s"
                % (filePath, len(strings), len(opcodes), len(exports), imphash)
            )

    # return it as a set (unique strings)
    return all_strings, all_opcodes, all_imphashes, all_exports


def processSampleDir(targetDir, state):
    """
    Processes samples in a given directory and creates a yara rule file
    :param directory:
    :return:
    """

    # Extract all information
    (sample_string_stats, sample_opcode_stats, file_info, sample_utf16string_stats) = (
        parse_sample_dir(
            targetDir,
            state,
            state.args.nr,
            generateInfo=True,
            onlyRelevantExtensions=state.args.oe,
        )
    )

    # Evaluate Strings
    (file_strings, file_opcodes, combinations, super_rules, inverse_stats) = (
        sample_string_evaluation(
            sample_string_stats,
            sample_opcode_stats,
            file_info,
            state,
            sample_utf16string_stats,
        )
    )

    # Create Rule Files
    (rule_count, inverse_rule_count, super_rule_count) = generate_rules(
        state,
        file_strings,
        file_opcodes,
        super_rules,
        file_info,
        inverse_stats,
    )

    if state.args.inverse:
        print("[=] Generated %s INVERSE rules." % str(inverse_rule_count))
    else:
        print("[=] Generated %s SIMPLE rules." % str(rule_count))
        if not state.args.nosuper:
            print("[=] Generated %s SUPER rules." % str(super_rule_count))
        print("[=] All rules written to %s" % state.args.o)




from app.parse_files import extract_strings 

def test_string_extraction():  
    strings, utf16strs = extract_strings(b"string1\0string2\nmultilinestring\n1\0string1") 

    assert strings["string1"].count == 2
    assert strings["string2"].count == 1
    assert strings["multilinestring"].count == 1


import binascii
import datetime
import gzip
from typing import Tuple
import orjson as json
import os
import re
import sys
import traceback
import lief
from lxml import etree

import yargen_rs

PE_STRINGS_FILE = "./3rdparty/strings.xml"


def extract_opcodes(fileData) -> list[str]:
    return yargen_rs.extract_opcodes(fileData)


def get_pe_info(fileData: bytes) -> tuple[str, list[str]]:
    """
    Get different PE attributes and hashes by lief
    :param fileData:
    :return:
    """
    imphash = ""
    exports = []
    # Check for MZ header (speed improvement)
    if fileData[:2] != b"MZ":
        return imphash, exports
    try:

        print("Extracting PE information")
        binary = lief.parse(fileData)
        # Imphash
        imphash = lief.PE.get_imphash(binary, lief.PE.IMPHASH_MODE.PEFILE)
        # Exports (names)
        for exp in binary.get_export().entries:
            exports.append(str(exp.name))
    except Exception as e:
        traceback.print_exc()
        pass

    return imphash, exports


def get_abs_path(filename):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), filename)


def initialize_pestudio_strings():
    if not os.path.isfile(get_abs_path(PE_STRINGS_FILE)):
        return None
    print("[+] Processing PEStudio strings ...")

    pestudio_strings = {}

    tree = etree.parse(get_abs_path(PE_STRINGS_FILE))

    pestudio_strings["strings"] = tree.findall(".//string")
    pestudio_strings["av"] = tree.findall(".//av")
    pestudio_strings["folder"] = tree.findall(".//folder")
    pestudio_strings["os"] = tree.findall(".//os")
    pestudio_strings["reg"] = tree.findall(".//reg")
    pestudio_strings["guid"] = tree.findall(".//guid")
    pestudio_strings["ssdl"] = tree.findall(".//ssdl")
    pestudio_strings["ext"] = tree.findall(".//ext")
    pestudio_strings["agent"] = tree.findall(".//agent")
    pestudio_strings["oid"] = tree.findall(".//oid")
    pestudio_strings["priv"] = tree.findall(".//priv")

    # Obsolete
    # for elem in string_elems:
    #    strings.append(elem.text)

    return pestudio_strings


def emptyFolder(dir):
    """
    Removes all files from a given folder
    :return:
    """
    for file in os.listdir(dir):
        filePath = os.path.join(dir, file)
        try:
            if os.path.isfile(filePath):
                print("[!] Removing %s ..." % filePath)
                os.unlink(filePath)
        except Exception as e:
            print(e)


def getReference(ref):
    """
    Get a reference string - if the provided string is the path to a text file, then read the contents and return it as
    reference
    :param ref:
    :return:
    """
    if os.path.exists(ref):
        reference = getFileContent(ref)
        print("[+] Read reference from file %s > %s" % (ref, reference))
        return reference
    else:
        return ref


def save(object, filename):
    file = gzip.GzipFile(filename, "wb")
    file.write(bytes(json.dumps(object), "utf-8"))
    file.close()


def removeNonAsciiDrop(string):
    nonascii = "error"
    try:
        byte_list = [i.to_bytes(1, sys.byteorder) for i in string]
        # Generate a new string without disturbing characters
        nonascii = b"".join(i for i in byte_list if ord(i) < 127 and ord(i) > 31)
    except Exception as e:
        traceback.print_exc()
        pass
    return nonascii


def load(filename):
    file = gzip.GzipFile(filename, "rb")
    object = json.loads(file.read())
    file.close()
    return object


def getIdentifier(id, path):
    """
    Get a identifier string - if the provided string is the path to a text file, then read the contents and return it as
    reference, otherwise use the last element of the full path
    :param ref:
    :return:
    """
    # Identifier
    if id == "not set" or not os.path.exists(id):
        # Identifier is the highest folder name
        return os.path.basename(path.rstrip("/"))
    else:
        # Read identifier from file
        identifier = getFileContent(id)
        print("[+] Read identifier from file %s > %s" % (id, identifier))
        return identifier


def get_pestudio_score(string, pestudio_strings):
    for type in pestudio_strings:
        for elem in pestudio_strings[type]:
            # Full match
            if elem.text.lower() == string.lower():
                # Exclude the "extension" black list for now
                if type != "ext":
                    return 5, type
    return 0, ""


def getPrefix(prefix, identifier):
    """
    Get a prefix string for the rule description based on the identifier
    :param prefix:
    :param identifier:
    :return:
    """
    if prefix == "Auto-generated rule":
        return identifier
    else:
        return prefix


def getFileContent(file):
    """
    Gets the contents of a file (limited to 1024 characters)
    :param file:
    :return:
    """
    try:
        with open(file) as f:
            return f.read(1024)
    except Exception as e:
        return "not found"


def get_timestamp_basic(date_obj=None):
    if not date_obj:
        date_obj = datetime.datetime.now()
    date_str = date_obj.strftime("%Y-%m-%d")
    return date_str


def is_ascii_char(b, padding_allowed=False):
    if padding_allowed:
        if (ord(b) < 127 and ord(b) > 31) or ord(b) == 0:
            return 1
    else:
        if ord(b) < 127 and ord(b) > 31:
            return 1
    return 0


def is_ascii_string(string, padding_allowed=False):
    for b in [i.to_bytes(1, sys.byteorder) for i in string]:
        if padding_allowed:
            if not ((ord(b) < 127 and ord(b) > 31) or ord(b) == 0):
                return 0
        else:
            if not (ord(b) < 127 and ord(b) > 31):
                return 0
    return 1


def is_base_64(s):
    return (len(s) % 4 == 0) and re.match("^[A-Za-z0-9+/]+[=]{0,2}$", s)


def get_files(folder, notRecursive):
    # Not Recursive
    if notRecursive:
        for filename in os.listdir(folder):
            filePath = os.path.join(folder, filename)
            if os.path.isdir(filePath):
                continue
            yield filePath
    # Recursive
    else:
        for root, dirs, files in os.walk(folder, topdown=False):
            for name in files:
                filePath = os.path.join(root, name)
                yield filePath


def is_hex_encoded(s, check_length=True):
    if re.match("^[A-Fa-f0-9]+$", s):
        if check_length:
            if len(s) % 2 == 0:
                return True
        else:
            return True
    return False


def get_opcode_string(opcode):
    return " ".join(opcode[i : i + 2] for i in range(0, len(opcode), 2))


def get_uint_string(magic):
    if len(magic) == 2:
        return "uint8(0) == 0x{0}{1}".format(magic[0], magic[1])
    if len(magic) == 4:
        return "uint16(0) == 0x{2}{3}{0}{1}".format(
            magic[0], magic[1], magic[2], magic[3]
        )
    return ""



import os
import shutil
import sys
import traceback
import urllib


REPO_URLS = {
    "good-opcodes-part1.db": "https://www.bsk-consulting.de/yargen/good-opcodes-part1.db",
    "good-opcodes-part2.db": "https://www.bsk-consulting.de/yargen/good-opcodes-part2.db",
    "good-opcodes-part3.db": "https://www.bsk-consulting.de/yargen/good-opcodes-part3.db",
    "good-opcodes-part4.db": "https://www.bsk-consulting.de/yargen/good-opcodes-part4.db",
    "good-opcodes-part5.db": "https://www.bsk-consulting.de/yargen/good-opcodes-part5.db",
    "good-opcodes-part6.db": "https://www.bsk-consulting.de/yargen/good-opcodes-part6.db",
    "good-opcodes-part7.db": "https://www.bsk-consulting.de/yargen/good-opcodes-part7.db",
    "good-opcodes-part8.db": "https://www.bsk-consulting.de/yargen/good-opcodes-part8.db",
    "good-opcodes-part9.db": "https://www.bsk-consulting.de/yargen/good-opcodes-part9.db",
    "good-strings-part1.db": "https://www.bsk-consulting.de/yargen/good-strings-part1.db",
    "good-strings-part2.db": "https://www.bsk-consulting.de/yargen/good-strings-part2.db",
    "good-strings-part3.db": "https://www.bsk-consulting.de/yargen/good-strings-part3.db",
    "good-strings-part4.db": "https://www.bsk-consulting.de/yargen/good-strings-part4.db",
    "good-strings-part5.db": "https://www.bsk-consulting.de/yargen/good-strings-part5.db",
    "good-strings-part6.db": "https://www.bsk-consulting.de/yargen/good-strings-part6.db",
    "good-strings-part7.db": "https://www.bsk-consulting.de/yargen/good-strings-part7.db",
    "good-strings-part8.db": "https://www.bsk-consulting.de/yargen/good-strings-part8.db",
    "good-strings-part9.db": "https://www.bsk-consulting.de/yargen/good-strings-part9.db",
    "good-exports-part1.db": "https://www.bsk-consulting.de/yargen/good-exports-part1.db",
    "good-exports-part2.db": "https://www.bsk-consulting.de/yargen/good-exports-part2.db",
    "good-exports-part3.db": "https://www.bsk-consulting.de/yargen/good-exports-part3.db",
    "good-exports-part4.db": "https://www.bsk-consulting.de/yargen/good-exports-part4.db",
    "good-exports-part5.db": "https://www.bsk-consulting.de/yargen/good-exports-part5.db",
    "good-exports-part6.db": "https://www.bsk-consulting.de/yargen/good-exports-part6.db",
    "good-exports-part7.db": "https://www.bsk-consulting.de/yargen/good-exports-part7.db",
    "good-exports-part8.db": "https://www.bsk-consulting.de/yargen/good-exports-part8.db",
    "good-exports-part9.db": "https://www.bsk-consulting.de/yargen/good-exports-part9.db",
    "good-imphashes-part1.db": "https://www.bsk-consulting.de/yargen/good-imphashes-part1.db",
    "good-imphashes-part2.db": "https://www.bsk-consulting.de/yargen/good-imphashes-part2.db",
    "good-imphashes-part3.db": "https://www.bsk-consulting.de/yargen/good-imphashes-part3.db",
    "good-imphashes-part4.db": "https://www.bsk-consulting.de/yargen/good-imphashes-part4.db",
    "good-imphashes-part5.db": "https://www.bsk-consulting.de/yargen/good-imphashes-part5.db",
    "good-imphashes-part6.db": "https://www.bsk-consulting.de/yargen/good-imphashes-part6.db",
    "good-imphashes-part7.db": "https://www.bsk-consulting.de/yargen/good-imphashes-part7.db",
    "good-imphashes-part8.db": "https://www.bsk-consulting.de/yargen/good-imphashes-part8.db",
    "good-imphashes-part9.db": "https://www.bsk-consulting.de/yargen/good-imphashes-part9.db",
}


def update_databases(args):
    # Preparations
    try:
        dbDir = "./dbs/"
        if not os.path.exists(dbDir):
            os.makedirs(dbDir)
    except Exception as e:
        if args.debug:
            traceback.print_exc()
        print("Error while creating the database directory ./dbs")
        sys.exit(1)

    # Downloading current repository
    try:
        for filename, repo_url in REPO_URLS.items():
            print("Downloading %s from %s ..." % (filename, repo_url))
            with urllib.request.urlopen(repo_url) as response, open(
                "./dbs/%s" % filename, "wb"
            ) as out_file:
                shutil.copyfileobj(response, out_file)
    except Exception as e:
        if args.debug:
            traceback.print_exc()
        print(
            "Error while downloading the database file - check your Internet connection "
            "(try to run it with --debug to see the full error message)"
        )
        sys.exit(1)



from collections import Counter
import traceback

from app.scoring import filter_opcode_set, filter_string_set
from app.utils import *

KNOWN_IMPHASHES = {
    "a04dd9f5ee88d7774203e0a0cfa1b941": "PsExec",
    "2b8c9d9ab6fefc247adaf927e83dcea6": "RAR SFX variant",
}


AI_COMMENT = """
The provided rule is a YARA rule, encompassing a wide range of suspicious strings. Kindly review the list and pinpoint the twenty strings that are most distinctive or appear most suited for a YARA rule focused on malware detection. Arrange them in descending order based on their level of suspicion. Then, swap out the current list of strings in the YARA rule with your chosen set and supply the revised rule.
---
"""


def get_file_range(size, fm_size):
    size_string = ""
    try:
        # max sample size - args.fm times the original size
        max_size_b = size * fm_size
        # Minimum size
        if max_size_b < 1024:
            max_size_b = 1024
        # in KB
        max_size = int(max_size_b / 1024)
        max_size_kb = max_size
        # Round
        if len(str(max_size)) == 2:
            max_size = int(round(max_size, -1))
        elif len(str(max_size)) == 3:
            max_size = int(round(max_size, -2))
        elif len(str(max_size)) == 4:
            max_size = int(round(max_size, -3))
        elif len(str(max_size)) >= 5:
            max_size = int(round(max_size, -3))
        size_string = "filesize < {0}KB".format(max_size)
        print(
            "File Size Eval: SampleSize (b): {0} SizeWithMultiplier (b/Kb): {1} / {2} RoundedSize: {3}".format(
                str(size), str(max_size_b), str(max_size_kb), str(max_size)
            )
        )
    except Exception:
        traceback.print_exc()
    finally:
        return size_string


def get_strings(state, string_elements):
    """
    Get a dictionary of all string types
    :param string_elements:
    :return:
    """
    strings = {
        "ascii": [],
        "wide": [],
        "base64 encoded": [],
        "hex encoded": [],
        "reversed": [],
    }

    # Adding the strings --------------------------------------
    if state.args.debug:
        print(
            state.base64strings,
            state.hexEncStrings,
            state.reversedStrings,
            state.utf16strings,
        )
    for i, string in enumerate(string_elements):

        if string[:8] == "UTF16LE:":
            string = string[8:]
            strings["wide"].append(string)
        if string in state.utf16strings:
            strings["wide"].append(string)
        elif string in state.base64strings:
            strings["base64 encoded"].append(string)
        elif string in state.hexEncStrings:
            strings["hex encoded"].append(string)
        elif string in state.reversedStrings:
            strings["reversed"].append(string)
        else:
            strings["ascii"].append(string)

    return strings


def write_strings(filePath, strings, output_dir, scores, stringScores):
    """
    Writes string information to an output file
    :param filePath:
    :param strings:
    :param output_dir:
    :param scores:
    :return:
    """
    SECTIONS = ["ascii", "wide", "base64 encoded", "hex encoded", "reversed"]
    # File
    filename = os.path.basename(filePath)
    strings_filename = os.path.join(output_dir, "%s_strings.txt" % filename)
    print("[+] Writing strings to file %s" % strings_filename)
    # Strings
    output_string = []
    for key in SECTIONS:
        # Skip empty
        if len(strings[key]) < 1:
            continue
        # Section
        output_string.append("%s Strings" % key.upper())
        output_string.append(
            "------------------------------------------------------------------------"
        )
        for string in strings[key]:
            if scores:
                score = "unknown"
                if key == "wide":
                    score = stringScores["UTF16LE:%s" % string]
                else:
                    score = stringScores[string]
                output_string.append("%d;%s" % (score, string))
            else:
                output_string.append(string)
        # Empty line between sections
        output_string.append("\n")
    with open(strings_filename, "w") as fh:
        fh.write("\n".join(output_string))


def generate_general_condition(state, file_info):
    """
    Generates a general condition for a set of files
    :param file_info:
    :return:
    """
    conditions_string = ""
    conditions = []
    pe_module_neccessary = False

    # Different Magic Headers and File Sizes
    magic_headers = []
    file_sizes = []
    imphashes = []

    try:
        for filePath in file_info:
            # Short file name info used for inverse generation has no magic/size fields
            if "magic" not in file_info[filePath]:
                continue
            magic = file_info[filePath]["magic"]
            size = file_info[filePath]["size"]
            imphash = file_info[filePath]["imphash"]

            # Add them to the lists
            if magic not in magic_headers and magic != "":
                magic_headers.append(magic)
            if size not in file_sizes:
                file_sizes.append(size)
            if imphash not in imphashes and imphash != "":
                imphashes.append(imphash)

        # If different magic headers are less than 5
        if len(magic_headers) <= 5:
            magic_string = " or ".join(get_uint_string(h) for h in magic_headers)
            if " or " in magic_string:
                conditions.append("( {0} )".format(magic_string))
            else:
                conditions.append("{0}".format(magic_string))

        # Biggest size multiplied with maxsize_multiplier
        if not state.args.nofilesize and len(file_sizes) > 0:
            conditions.append(get_file_range(max(file_sizes), state.args.fm))

        # If different magic headers are less than 5
        if len(imphashes) == 1:
            conditions.append('pe.imphash() == "{0}"'.format(imphashes[0]))
            pe_module_neccessary = True

        # If enough attributes were special
        condition_string = " and ".join(conditions)

    except Exception as e:
        if state.args.debug:
            traceback.print_exc()
            exit(1)
        print(
            "[E] ERROR while generating general condition - check the global rule and remove it if it's faulty"
        )

    return condition_string, pe_module_neccessary


def generate_rules(
    state,
    file_strings,
    file_opcodes,
    super_rules,
    file_info,
    inverse_stats,
):
    # Write to file ---------------------------------------------------
    if state.args.o:
        try:
            fh = open(state.args.o, "w")
        except Exception as e:
            traceback.print_exc()

    # General Info
    general_info = "/*\n"
    general_info += "   YARA Rule Set\n"
    general_info += "   Author: {0}\n".format(state.args.a)
    general_info += "   Date: {0}\n".format(get_timestamp_basic())
    general_info += "   Identifier: {0}\n".format(state.args.identifier)
    general_info += "   Reference: {0}\n".format(state.args.reference)
    if state.args.l != "":
        general_info += "   License: {0}\n".format(state.args.l)
    general_info += "*/\n\n"

    fh.write(general_info)

    # GLOBAL RULES ----------------------------------------------------
    if state.args.globalrule:

        condition, pe_module_necessary = generate_general_condition(file_info)

        # Global Rule
        if condition != "":
            global_rule = "/* Global Rule -------------------------------------------------------------- */\n"
            global_rule += "/* Will be evaluated first, speeds up scanning process, remove at will */\n\n"
            global_rule += "global private rule gen_characteristics {\n"
            global_rule += "   condition:\n"
            global_rule += "      {0}\n".format(condition)
            global_rule += "}\n\n"

            # Write rule
            if state.args.o:
                fh.write(global_rule)

    # General vars
    rules = ""
    printed_rules = {}
    opcodes_to_add = []
    rule_count = 0
    inverse_rule_count = 0
    super_rule_count = 0
    pe_module_necessary = False

    if not state.args.inverse:
        # PROCESS SIMPLE RULES ----------------------------------------------------
        print("[+] Generating Simple Rules ...")
        # Apply intelligent filters
        print("[-] Applying intelligent filters to string findings ...")
        for filePath in file_strings:

            print("[-] Filtering string set for %s ..." % filePath)

            # Replace the original string set with the filtered one
            file_strings[filePath] = filter_string_set(file_strings[filePath], state)

            print("[-] Filtering opcode set for %s ..." % filePath)

            # Replace the original opcode set with the filtered one
            file_opcodes[filePath] = (
                filter_opcode_set(state, file_opcodes[filePath], state.good_opcodes_db)
                if filePath in file_opcodes
                else []
            )

        # GENERATE SIMPLE RULES -------------------------------------------
        fh.write(
            "/* Rule Set ----------------------------------------------------------------- */\n\n"
        )

        for filePath in file_strings:

            # Skip if there is nothing to do
            if len(file_strings[filePath]) == 0:
                print(
                    "[W] Not enough high scoring strings to create a rule. "
                    "(Try -z 0 to reduce the min score or --opcodes to include opcodes) FILE: %s"
                    % filePath
                )
                continue
            elif len(file_strings[filePath]) == 0 and len(file_opcodes[filePath]) == 0:
                print(
                    "[W] Not enough high scoring strings and opcodes to create a rule. "
                    "(Try -z 0 to reduce the min score) FILE: %s" % filePath
                )
                continue

            # Create Rule
            try:
                rule = ""
                (path, file) = os.path.split(filePath)
                # Prepare name
                fileBase = os.path.splitext(file)[0]
                # Create a clean new name
                cleanedName = fileBase
                # Adapt length of rule name
                if len(fileBase) < 8:  # if name is too short add part from path
                    cleanedName = path.split("\\")[-1:][0] + "_" + cleanedName
                # File name starts with a number
                if re.search(r"^[0-9]", cleanedName):
                    cleanedName = "sig_" + cleanedName
                # clean name from all characters that would cause errors
                cleanedName = re.sub(r"[^\w]", "_", cleanedName)
                # Check if already printed
                if cleanedName in printed_rules:
                    printed_rules[cleanedName] += 1
                    cleanedName = cleanedName + "_" + str(printed_rules[cleanedName])
                else:
                    printed_rules[cleanedName] = 1

                # Print rule title ----------------------------------------
                rule += "rule %s {\n" % cleanedName

                # Meta data -----------------------------------------------
                rule += "   meta:\n"
                rule += '      description = "%s - file %s"\n' % (
                    state.args.prefix,
                    file,
                )
                rule += '      author = "%s"\n' % state.args.a
                rule += '      reference = "%s"\n' % state.args.reference
                rule += '      date = "%s"\n' % get_timestamp_basic()
                rule += '      hash1 = "%s"\n' % file_info[filePath]["hash"]
                rule += "   strings:\n"

                # Get the strings -----------------------------------------
                # Rule String generation
                (
                    rule_strings,
                    opcodes_included,
                    string_rule_count,
                    high_scoring_strings,
                ) = get_rule_strings(
                    state,
                    file_strings[filePath],
                    file_opcodes[filePath],
                )

                rule += rule_strings

                # Extract rul strings
                if state.args.strings:
                    strings = get_strings(
                        state,
                        file_strings[filePath],
                    )
                    write_strings(
                        filePath,
                        strings,
                        state.args.e,
                        state.args.score,
                        state.stringScores,
                    )

                # Condition -----------------------------------------------
                # Conditions list (will later be joined with 'or')
                conditions = []  # AND connected
                subconditions = []  # OR connected

                # Condition PE
                # Imphash and Exports - applicable to PE files only
                condition_pe = []
                condition_pe_part1 = []
                condition_pe_part2 = []
                if not state.args.noextras and file_info[filePath]["magic"] == "MZ":
                    # Add imphash - if certain conditions are met
                    if (
                        file_info[filePath]["imphash"] not in state.good_imphashes_db
                        and file_info[filePath]["imphash"] != ""
                    ):
                        # Comment to imphash
                        imphash = file_info[filePath]["imphash"]
                        comment = ""
                        if imphash in KNOWN_IMPHASHES:
                            comment = " /* {0} */".format(KNOWN_IMPHASHES[imphash])
                        # Add imphash to condition
                        condition_pe_part1.append(
                            'pe.imphash() == "{0}"{1}'.format(imphash, comment)
                        )
                        pe_module_necessary = True
                    if file_info[filePath]["exports"]:
                        e_count = 0
                        for export in file_info[filePath]["exports"]:
                            if export not in state.good_exports_db:
                                condition_pe_part2.append(
                                    'pe.exports("{0}")'.format(export)
                                )
                                e_count += 1
                                pe_module_necessary = True
                            if e_count > 5:
                                break

                # 1st Part of Condition 1
                basic_conditions = []
                # Filesize
                if not state.args.nofilesize:
                    basic_conditions.insert(
                        0, get_file_range(file_info[filePath]["size"], state.args.fm)
                    )
                # Magic
                if file_info[filePath]["magic"] != "":
                    uint_string = get_uint_string(file_info[filePath]["magic"])
                    basic_conditions.insert(0, uint_string)
                # Basic Condition
                if len(basic_conditions):
                    conditions.append(" and ".join(basic_conditions))

                # Add extra PE conditions to condition 1
                pe_conditions_add = False
                if condition_pe_part1 or condition_pe_part2:
                    if len(condition_pe_part1) == 1:
                        condition_pe.append(condition_pe_part1[0])
                    elif len(condition_pe_part1) > 1:
                        condition_pe.append("( %s )" % " or ".join(condition_pe_part1))
                    if len(condition_pe_part2) == 1:
                        condition_pe.append(condition_pe_part2[0])
                    elif len(condition_pe_part2) > 1:
                        condition_pe.append("( %s )" % " and ".join(condition_pe_part2))
                    # Marker that PE conditions have been added
                    pe_conditions_add = True
                    # Add to sub condition
                    subconditions.append(" and ".join(condition_pe))

                # String combinations
                cond_op = ""  # opcodes condition
                cond_hs = ""  # high scoring strings condition
                cond_ls = ""  # low scoring strings condition

                low_scoring_strings = string_rule_count - high_scoring_strings
                if high_scoring_strings > 0:
                    cond_hs = "1 of ($x*)"
                if low_scoring_strings > 0:
                    if low_scoring_strings > 10:
                        if high_scoring_strings > 0:
                            cond_ls = "4 of them"
                        else:
                            cond_ls = "8 of them"
                    else:
                        cond_ls = "all of them"

                # If low scoring and high scoring
                cond_combined = "all of them"
                needs_brackets = False
                if low_scoring_strings > 0 and high_scoring_strings > 0:
                    # If PE conditions have been added, don't be so strict with the strings
                    if pe_conditions_add:
                        cond_combined = "{0} or {1}".format(cond_hs, cond_ls)
                        needs_brackets = True
                    else:
                        cond_combined = "{0} and {1}".format(cond_hs, cond_ls)
                elif low_scoring_strings > 0 and not high_scoring_strings > 0:
                    cond_combined = "{0}".format(cond_ls)
                elif not low_scoring_strings > 0 and high_scoring_strings > 0:
                    cond_combined = "{0}".format(cond_hs)
                if opcodes_included:
                    cond_op = " and all of ($op*)"

                # Opcodes (if needed)
                if cond_op or needs_brackets:
                    subconditions.append("( {0}{1} )".format(cond_combined, cond_op))
                else:
                    subconditions.append(cond_combined)

                # Now add string condition to the conditions
                if len(subconditions) == 1:
                    conditions.append(subconditions[0])
                elif len(subconditions) > 1:
                    conditions.append("( %s )" % " or ".join(subconditions))

                # Create condition string
                condition_string = " and\n      ".join(conditions)

                rule += "   condition:\n"
                rule += "      %s\n" % condition_string
                rule += "}\n\n"

                # Add to rules string
                rules += rule

                rule_count += 1
            except Exception as e:
                traceback.print_exc()

    # GENERATE SUPER RULES --------------------------------------------
    if not state.args.nosuper and not state.args.inverse:

        rules += "/* Super Rules ------------------------------------------------------------- */\n\n"
        super_rule_names = []

        print("[+] Generating Super Rules ...")
        printed_combi = {}
        for super_rule in super_rules:
            try:
                rule = ""
                # Prepare Name
                rule_name = ""
                file_list = []

                # Loop through files
                imphashes = Counter()
                for filePath in super_rule["files"]:
                    (path, file) = os.path.split(filePath)
                    file_list.append(file)
                    # Prepare name
                    fileBase = os.path.splitext(file)[0]
                    # Create a clean new name
                    cleanedName = fileBase
                    # Append it to the full name
                    rule_name += "_" + cleanedName
                    # Check if imphash of all files is equal
                    imphash = file_info[filePath]["imphash"]
                    if imphash != "-" and imphash != "":
                        imphashes.update([imphash])

                # Imphash usable
                if len(imphashes) == 1:
                    unique_imphash = list(imphashes.items())[0][0]
                    if unique_imphash in state.good_imphashes_db:
                        unique_imphash = ""

                # Shorten rule name
                rule_name = rule_name[:124]
                # Add count if rule name already taken
                if rule_name not in super_rule_names:
                    rule_name = "%s_%s" % (rule_name, super_rule_count)
                super_rule_names.append(rule_name)

                # Create a list of files
                file_listing = ", ".join(file_list)

                # File name starts with a number
                if re.search(r"^[0-9]", rule_name):
                    rule_name = "sig_" + rule_name
                # clean name from all characters that would cause errors
                rule_name = re.sub(r"[^\w]", "_", rule_name)
                # Check if already printed
                if rule_name in printed_rules:
                    printed_combi[rule_name] += 1
                    rule_name = rule_name + "_" + str(printed_combi[rule_name])
                else:
                    printed_combi[rule_name] = 1

                # Print rule title
                rule += "rule %s {\n" % rule_name
                rule += "   meta:\n"
                rule += '      description = "%s - from files %s"\n' % (
                    state.args.prefix,
                    file_listing,
                )
                rule += '      author = "%s"\n' % state.args.a
                rule += '      reference = "%s"\n' % state.args.reference
                rule += '      date = "%s"\n' % get_timestamp_basic()
                for i, filePath in enumerate(super_rule["files"]):
                    rule += '      hash%s = "%s"\n' % (
                        str(i + 1),
                        file_info[filePath]["hash"],
                    )

                rule += "   strings:\n"

                # Adding the opcodes
                if file_opcodes.get(filePath) is None:
                    tmp_file_opcodes = {}
                else:
                    tmp_file_opcodes = file_opcodes.get(filePath)
                (
                    rule_strings,
                    opcodes_included,
                    string_rule_count,
                    high_scoring_strings,
                ) = get_rule_strings(
                    state,
                    super_rule["strings"],
                    tmp_file_opcodes,
                )
                rule += rule_strings

                # Condition -----------------------------------------------
                # Conditions list (will later be joined with 'or')
                conditions = []

                # 1st condition
                # Evaluate the general characteristics
                file_info_super = {}
                for filePath in super_rule["files"]:
                    file_info_super[filePath] = file_info[filePath]
                condition_strings, pe_module_necessary_gen = generate_general_condition(
                    state, file_info_super
                )
                if pe_module_necessary_gen:
                    pe_module_necessary = True

                # 2nd condition
                # String combinations
                cond_op = ""  # opcodes condition
                cond_hs = ""  # high scoring strings condition
                cond_ls = ""  # low scoring strings condition

                low_scoring_strings = string_rule_count - high_scoring_strings
                if high_scoring_strings > 0:
                    cond_hs = "1 of ($x*)"
                if low_scoring_strings > 0:
                    if low_scoring_strings > 10:
                        if high_scoring_strings > 0:
                            cond_ls = "4 of them"
                        else:
                            cond_ls = "8 of them"
                    else:
                        cond_ls = "all of them"

                # If low scoring and high scoring
                cond_combined = "all of them"
                if low_scoring_strings > 0 and high_scoring_strings > 0:
                    cond_combined = "{0} and {1}".format(cond_hs, cond_ls)
                elif low_scoring_strings > 0 and not high_scoring_strings > 0:
                    cond_combined = "{0}".format(cond_ls)
                elif not low_scoring_strings > 0 and high_scoring_strings > 0:
                    cond_combined = "{0}".format(cond_hs)
                if opcodes_included:
                    cond_op = " and all of ($op*)"

                condition2 = "( {0} ){1}".format(cond_combined, cond_op)
                conditions.append(" and ".join([condition_strings, condition2]))

                # 3nd condition
                # In memory detection base condition (no magic, no filesize)
                condition_pe = "all of them"
                conditions.append(condition_pe)

                # Create condition string
                condition_string = "\n      ) or ( ".join(conditions)

                rule += "   condition:\n"
                rule += "      ( %s )\n" % condition_string
                rule += "}\n\n"

                # print rule
                # Add to rules string
                rules += rule

                super_rule_count += 1
            except Exception as e:
                traceback.print_exc()

    try:
        # WRITING RULES TO FILE
        # PE Module -------------------------------------------------------
        if not state.args.noextras:
            if pe_module_necessary:
                fh.write('import "pe"\n\n')
        # RULES -----------------------------------------------------------
        if state.args.o:
            fh.write(rules)
    except Exception as e:
        traceback.print_exc()

    # PROCESS INVERSE RULES ---------------------------------------------------
    # print inverse_stats.keys()
    if state.args.inverse:
        print("[+] Generating inverse rules ...")
        inverse_rules = ""
        # Apply intelligent filters -------------------------------------------
        print("[+] Applying intelligent filters to string findings ...")
        for fileName in inverse_stats:

            print("[-] Filtering string set for %s ..." % fileName)

            # Replace the original string set with the filtered one
            string_set = inverse_stats[fileName]
            inverse_stats[fileName] = []
            inverse_stats[fileName] = filter_string_set(string_set, state)

            # Preset if empty
            if fileName not in file_opcodes:
                file_opcodes[fileName] = {}

        # GENERATE INVERSE RULES -------------------------------------------
        fh.write(
            "/* Inverse Rules ------------------------------------------------------------- */\n\n"
        )

        for fileName in inverse_stats:
            try:
                rule = ""
                # Create a clean new name
                cleanedName = fileName.replace(".", "_")
                # Add ANOMALY
                cleanedName += "_ANOMALY"
                # File name starts with a number
                if re.search(r"^[0-9]", cleanedName):
                    cleanedName = "sig_" + cleanedName
                # clean name from all characters that would cause errors
                cleanedName = re.sub(r"[^\w]", "_", cleanedName)
                # Check if already printed
                if cleanedName in printed_rules:
                    printed_rules[cleanedName] += 1
                    cleanedName = cleanedName + "_" + str(printed_rules[cleanedName])
                else:
                    printed_rules[cleanedName] = 1

                # Print rule title ----------------------------------------
                rule += "rule %s {\n" % cleanedName

                # Meta data -----------------------------------------------
                rule += "   meta:\n"
                rule += '      description = "%s for anomaly detection - file %s"\n' % (
                    state.args.prefix,
                    fileName,
                )
                rule += '      author = "%s"\n' % state.args.a
                rule += '      reference = "%s"\n' % state.args.reference
                rule += '      date = "%s"\n' % get_timestamp_basic()
                for i, hash in enumerate(file_info[fileName]["hashes"]):
                    rule += '      hash%s = "%s"\n' % (str(i + 1), hash)

                rule += "   strings:\n"

                # Get the strings -----------------------------------------
                # Rule String generation
                (
                    rule_strings,
                    opcodes_included,
                    string_rule_count,
                    high_scoring_strings,
                ) = get_rule_strings(
                    state,
                    inverse_stats[fileName],
                    file_opcodes[fileName],
                )

                rule += rule_strings

                # Condition -----------------------------------------------
                folderNames = ""
                if not state.args.nodirname:
                    folderNames += "and ( filepath matches /"
                    folderNames += "$/ or filepath matches /".join(
                        file_info[fileName]["folder_names"]
                    )
                    folderNames += "$/ )"
                condition = 'filename == "%s" %s and not ( all of them )' % (
                    fileName,
                    folderNames,
                )

                rule += "   condition:\n"
                rule += "      %s\n" % condition
                rule += "}\n\n"

                # print rule
                # Add to rules string
                inverse_rules += rule

            except Exception as e:
                traceback.print_exc()

        try:
            # Try to write rule to file
            if state.args.o:
                fh.write(inverse_rules)
            inverse_rule_count += 1
        except Exception as e:
            traceback.print_exc()

    # Close the rules file --------------------------------------------
    if state.args.o:
        try:
            fh.close()
        except Exception as e:
            traceback.print_exc()

    # Print rules to command line -------------------------------------
    if state.args.debug:
        print(rules)

    return (rule_count, inverse_rule_count, super_rule_count)


def get_rule_strings(state, string_elements, opcode_elements):
    rule_strings = ""
    high_scoring_strings = 0
    string_rule_count = 0

    # Adding the strings --------------------------------------

    string_elements = list(set(string_elements))

    string_elements = sorted(
        string_elements, key=lambda x: state.stringScores[x], reverse=True
    )
    for i, string in enumerate(string_elements):

        # Collect the data
        is_fullword = True
        initial_string = string
        enc = " ascii"
        base64comment = ""
        hexEncComment = ""
        reversedComment = ""
        fullword = ""
        pestudio_comment = ""
        score_comment = ""
        goodware_comment = ""

        if string in state.good_strings_db:
            goodware_comment = " /* Goodware String - occured %s times */" % (
                state.good_strings_db[string]
            )
        if string[:8] == "UTF16LE:":
            string = string[8:]
            enc = " wide"
        if string in state.stringScores:
            if state.args.score:
                cat_comment = state.string_to_comms[string]
                score_comment += (
                    f" /* score: {state.stringScores[string]}  {cat_comment}*/"
                )
        else:
            print("NO SCORE: %s" % string)

        if string in state.utf16strings:
            enc = " wide"
        if string in state.base64strings:
            base64comment = (
                " /* base64 encoded string '%s' */"
                % state.base64strings[string].decode()
            )
        if string in state.hexEncStrings:
            hexEncComment = (
                " /* hex encoded string '%s' */"
                % removeNonAsciiDrop(state.hexEncStrings[string]).decode()
            )
        if string in state.pestudioMarker and state.args.score:
            pestudio_comment = (
                " /* PEStudio Blacklist: %s */" % state.pestudioMarker[string]
            )
        if string in state.reversedStrings:
            reversedComment = (
                " /* reversed goodware string '%s' */" % state.reversedStrings[string]
            )

        # Extra checks
        if is_hex_encoded(string, check_length=False):
            is_fullword = False

        # Checking string length
        if len(string) >= state.args.s:
            # cut string
            string = string[: state.args.s].rstrip("\\")
            # not fullword anymore
            is_fullword = False
        # Show as fullword
        if is_fullword:
            fullword = " fullword"

        # Now compose the rule line
        if float(state.stringScores[initial_string]) > state.args.score_highly_specific:
            high_scoring_strings += 1
            rule_strings += '      $x%s = "%s"%s%s%s%s%s%s%s%s\n' % (
                str(i + 1),
                string,
                fullword,
                enc,
                base64comment,
                reversedComment,
                pestudio_comment,
                score_comment,
                goodware_comment,
                hexEncComment,
            )
        else:
            rule_strings += '      $s%s = "%s"%s%s%s%s%s%s%s%s\n' % (
                str(i + 1),
                string,
                fullword,
                enc,
                base64comment,
                reversedComment,
                pestudio_comment,
                score_comment,
                goodware_comment,
                hexEncComment,
            )

        # If too many string definitions found - cut it at the
        # count defined via command line param -rc
        if (i + 1) >= 20:  # state.args.strings_per_rule:
            break

        string_rule_count += 1

    # Adding the opcodes --------------------------------------
    opcodes_included = False
    if len(opcode_elements) > 0:
        rule_strings += "\n"
        for i, opcode in enumerate(opcode_elements):
            rule_strings += "      $op%s = { %s }\n" % (str(i), opcode)
            opcodes_included = True
    else:
        if state.args.opcodes:
            print("[-] Not enough unique opcodes found to include them")

    return rule_strings, opcodes_included, string_rule_count, high_scoring_strings



import argparse
import os
import sys

import dbs


def get_args():
    # Parse Arguments
    parser = argparse.ArgumentParser(description="yarGen")

    group_creation = parser.add_argument_group("Rule Creation")
    group_creation.add_argument("-m", help="Path to scan for malware")
    group_creation.add_argument(
        "-y",
        help="Minimum string length to consider (default=8)",
        metavar="min-size",
        default=8,
    )
    group_creation.add_argument(
        "-z",
        help="Minimum score to consider (default=0)",
        metavar="min-score",
        default=5,
    )
    group_creation.add_argument(
        "-x",
        help="Score required to set string as 'highly specific string' (default: 30)",
        metavar="high-scoring",
        default=30,
    )
    group_creation.add_argument(
        "-w",
        help="Minimum number of strings that overlap to create a super rule (default: 5)",
        metavar="superrule-overlap",
        default=5,
    )
    group_creation.add_argument(
        "-s",
        help="Maximum length to consider (default=128)",
        metavar="max-size",
        default=128,
        type=int,
    )
    group_creation.add_argument(
        "-rc",
        help="Maximum number of strings per rule (default=20, intelligent filtering "
        "will be applied)",
        metavar="strings-per-rule",
        default=20,
    )
    group_creation.add_argument(
        "--excludegood",
        help="Force the exclude all goodware strings",
        action="store_true",
        default=False,
    )

    group_output = parser.add_argument_group("Rule Output")
    group_output.add_argument(
        "-o",
        help="Output rule file",
        metavar="output_rule_file",
        default="yargen_rules.yar",
    )
    group_output.add_argument(
        "-e",
        help="Output directory for string exports",
        metavar="output_dir_strings",
        default="",
    )
    group_output.add_argument(
        "-a", help="Author Name", metavar="author", default="yarGen Rule Generator"
    )
    group_output.add_argument(
        "-r",
        help="Reference (can be string or text file)",
        metavar="ref",
        default="https://github.com/Neo23x0/yarGen",
    )
    group_output.add_argument("-l", help="License", metavar="lic", default="")
    group_output.add_argument(
        "-p",
        help="Prefix for the rule description",
        metavar="prefix",
        default="Auto-generated rule",
    )
    group_output.add_argument(
        "-b",
        help="Text file from which the identifier is read (default: last folder name in "
        'the full path, e.g. "myRAT" if -m points to /mnt/mal/myRAT)',
        metavar="identifier",
        default="not set",
    )
    group_output.add_argument(
        "--score",
        help="Show the string scores as comments in the rules",
        action="store_true",
        default=False,
    )
    group_output.add_argument(
        "--strings",
        help="Show the string scores as comments in the rules",
        action="store_true",
        default=False,
    )
    group_output.add_argument(
        "--nosimple",
        help="Skip simple rule creation for files included in super rules",
        action="store_true",
        default=False,
    )
    group_output.add_argument(
        "--nomagic",
        help="Don't include the magic header condition statement",
        action="store_true",
        default=False,
    )
    group_output.add_argument(
        "--nofilesize",
        help="Don't include the filesize condition statement",
        action="store_true",
        default=False,
    )
    group_output.add_argument(
        "-fm",
        help="Multiplier for the maximum 'filesize' condition value (default: 3)",
        default=3,
    )
    group_output.add_argument(
        "--globalrule",
        help="Create global rules (improved rule set speed)",
        action="store_true",
        default=False,
    )
    group_output.add_argument(
        "--nosuper",
        action="store_true",
        default=False,
        help="Don't try to create super rules " "that match against various files",
    )

    group_db = parser.add_argument_group("Database Operations")
    group_db.add_argument(
        "--update",
        action="store_true",
        default=False,
        help="Update the local strings and opcodes " "dbs from the online repository",
    )
    group_db.add_argument(
        "-g",
        help="Path to scan for goodware (dont use the database shipped with yaraGen)",
    )
    group_db.add_argument(
        "-u",
        action="store_true",
        default=False,
        help="Update local standard goodware database with "
        "a new analysis result (used with -g)",
    )
    group_db.add_argument(
        "-c",
        action="store_true",
        default=False,
        help="Create new local goodware database "
        '(use with -g and optionally -i "identifier")',
    )
    group_db.add_argument(
        "-i",
        default="",
        help="Specify an identifier for the newly created databases "
        "(good-strings-identifier.db, good-opcodes-identifier.db)",
    )

    group_general = parser.add_argument_group("General Options")
    group_general.add_argument(
        "--dropzone",
        action="store_true",
        default=False,
        help="Dropzone mode - monitors a directory [-m] for new samples to process. "
        "WARNING: Processed files will be deleted!",
    )
    group_general.add_argument(
        "--nr",
        action="store_true",
        default=False,
        help="Do not recursively scan directories",
    )
    group_general.add_argument(
        "--oe",
        action="store_true",
        default=False,
        help="Only scan executable extensions EXE, "
        "DLL, ASP, JSP, PHP, BIN, INFECTED",
    )
    group_general.add_argument(
        "-fs",
        help="Max file size in MB to analyze (default=10)",
        metavar="size-in-MB",
        default=10,
    )
    group_general.add_argument(
        "--noextras",
        action="store_true",
        default=False,
        help="Don't use extras like Imphash or PE header specifics",
    )
    group_general.add_argument(
        "--debug", action="store_true", default=False, help="Debug output"
    )
    group_general.add_argument(
        "--trace", action="store_true", default=False, help="Trace output"
    )

    group_opcode = parser.add_argument_group("Other Features")
    group_opcode.add_argument(
        "--opcodes",
        action="store_true",
        default=False,
        help="Do use the OpCode feature "
        "(use this if not enough high "
        "scoring strings can be found)",
    )
    group_opcode.add_argument(
        "-n",
        help="Number of opcodes to add if not enough high scoring string could be found "
        "(default=3)",
        metavar="opcode-num",
        default=3,
    )

    group_inverse = parser.add_argument_group("Inverse Mode (unstable)")
    group_inverse.add_argument(
        "--inverse", help=argparse.SUPPRESS, action="store_true", default=False
    )
    group_inverse.add_argument(
        "--nodirname", help=argparse.SUPPRESS, action="store_true", default=False
    )
    group_inverse.add_argument(
        "--noscorefilter", help=argparse.SUPPRESS, action="store_true", default=False
    )

    args = parser.parse_args()

    # Print Welcome

    if not args.update and not args.m and not args.g:
        parser.print_help()
        print("")
        print(
            """
[E] You have to select --update to update yarGens database or -m for signature generation or -g for the 
creation of goodware string collections 
(see https://github.com/Neo23x0/yarGen#examples for more details)

Recommended command line:
    python yarGen.py -a 'Your Name' --opcodes --dropzone -m ./dropzone"""
        )
        sys.exit(1)

    # Update
    if args.update:
        dbs.update_databases()
        print("[+] Updated databases - you can now start creating YARA rules")
        sys.exit(0)

    # Typical input erros
    if args.m:
        if os.path.isfile(args.m):
            print("[E] Input is a file, please use a directory instead (-m path)")
            sys.exit(0)

    return args
